@@ -16,49 +16,69 @@
 */
 
 #include "knotification.h"
+#include "debug_p.h"
 #include "knotification_p.h"
 #include "knotificationmanager_p.h"
 #include "knotificationreplyaction.h"
 
 #include <config-knotifications.h>
 
-#if HAVE_KWINDOWSYSTEM
-#include <KWindowSystem>
-#endif
 #include <QGuiApplication>
 
 #include <QStringList>
-#ifdef QT_WIDGETS_LIB
-#include <QTabWidget>
-#endif
 #include <QUrl>
 
 // incremental notification ID
 static int notificationIdCounter = 0;
 
-
-#if KNOTIFICATIONS_BUILD_DEPRECATED_SINCE(5, 75)
-KNotification::KNotification(const QString &eventId, QWidget *parent, const NotificationFlags &flags)
-    : QObject(parent)
-    , d(new Private)
+class KNotificationActionPrivate
 {
-    d->eventId = eventId;
-    d->flags = flags;
-    setWidget(parent);
-    connect(&d->updateTimer, &QTimer::timeout, this, &KNotification::update);
-    d->updateTimer.setSingleShot(true);
-    d->updateTimer.setInterval(100);
-    d->id = ++notificationIdCounter;
+public:
+    QString label;
+    QString id;
+};
 
-#if HAVE_KWINDOWSYSTEM
-    if (KWindowSystem::isPlatformWayland()) {
-        setHint(QStringLiteral("x-kde-xdgTokenAppId"), QGuiApplication::desktopFileName());
-    }
-#endif
+KNotificationAction::KNotificationAction(QObject *parent)
+    : QObject(parent)
+    , d(new KNotificationActionPrivate)
+{
 }
-#endif
 
-KNotification::KNotification(const QString &eventId, const NotificationFlags &flags, QObject *parent)
+KNotificationAction::KNotificationAction(const QString &label)
+    : QObject()
+    , d(new KNotificationActionPrivate)
+{
+    d->label = label;
+}
+
+KNotificationAction::~KNotificationAction()
+{
+}
+
+QString KNotificationAction::label() const
+{
+    return d->label;
+}
+
+void KNotificationAction::setLabel(const QString &label)
+{
+    if (d->label != label) {
+        d->label = label;
+        Q_EMIT labelChanged(label);
+    }
+}
+
+QString KNotificationAction::id() const
+{
+    return d->id;
+}
+
+void KNotificationAction::setId(const QString &id)
+{
+    d->id = id;
+}
+
+KNotification::KNotification(const QString &eventId, NotificationFlags flags, QObject *parent)
     : QObject(parent)
     , d(new Private)
 {
@@ -69,15 +89,18 @@ KNotification::KNotification(const QString &eventId, const NotificationFlags &fl
     d->updateTimer.setInterval(100);
     d->id = ++notificationIdCounter;
 
-#if HAVE_KWINDOWSYSTEM
-    if (KWindowSystem::isPlatformWayland()) {
+    if (QGuiApplication::platformName().startsWith(QLatin1String("wayland"))) {
         setHint(QStringLiteral("x-kde-xdgTokenAppId"), QGuiApplication::desktopFileName());
     }
-#endif
 }
 
 KNotification::~KNotification()
 {
+    if (d->ownsActions) {
+        qDeleteAll(d->actions);
+        delete d->defaultAction;
+    }
+
     if (d->id >= 0) {
         KNotificationManager::self()->close(d->id);
     }
@@ -106,22 +129,6 @@ QString KNotification::text() const
     return d->text;
 }
 
-QWidget *KNotification::widget() const
-{
-    return d->widget;
-}
-
-void KNotification::setWidget(QWidget *wid)
-{
-#ifdef QT_WIDGETS_LIB
-    d->widget = wid;
-    //     setParent(wid);
-    if (wid && d->flags & CloseWhenWidgetActivated) {
-        wid->installEventFilter(this);
-    }
-#endif
-}
-
 void KNotification::setTitle(const QString &title)
 {
     if (title == d->title) {
@@ -183,20 +190,64 @@ void KNotification::setPixmap(const QPixmap &pix)
     }
 }
 
-QStringList KNotification::actions() const
+QList<KNotificationAction *> KNotification::actions() const
 {
     return d->actions;
 }
 
-void KNotification::setActions(const QStringList &as)
+void KNotification::clearActions()
 {
-    if (as == d->actions) {
+    if (d->ownsActions) {
+        qDeleteAll(d->actions);
+    }
+    d->actions.clear();
+    d->actionIdCounter = 1;
+
+    d->needUpdate = true;
+    if (d->id >= 0) {
+        d->updateTimer.start();
+    }
+}
+
+KNotificationAction *KNotification::addAction(const QString &label)
+{
+    d->needUpdate = true;
+
+    KNotificationAction *action = new KNotificationAction(label);
+    action->setId(QString::number(d->actionIdCounter));
+    d->actionIdCounter++;
+
+    d->actions << action;
+    d->ownsActions = true;
+    Q_EMIT actionsChanged();
+
+    if (d->id >= 0) {
+        d->updateTimer.start();
+    }
+
+    return action;
+}
+
+void KNotification::setActionsQml(QList<KNotificationAction *> actions)
+{
+    if (actions == d->actions) {
         return;
     }
 
+    d->actions.clear();
+
     d->needUpdate = true;
-    d->actions = as;
+    d->actions = actions;
+    d->ownsActions = false;
     Q_EMIT actionsChanged();
+
+    int idCounter = 1;
+
+    for (KNotificationAction *action : d->actions) {
+        action->setId(QString::number(idCounter));
+        ++idCounter;
+    }
+
     if (d->id >= 0) {
         d->updateTimer.start();
     }
@@ -220,7 +271,27 @@ void KNotification::setReplyAction(std::unique_ptr<KNotificationReplyAction> rep
     }
 }
 
-void KNotification::setDefaultAction(const QString &defaultAction)
+KNotificationAction *KNotification::addDefaultAction(const QString &label)
+{
+    if (d->ownsActions) {
+        delete d->defaultAction;
+    }
+
+    d->needUpdate = true;
+    d->ownsActions = true;
+    d->defaultAction = new KNotificationAction(label);
+
+    d->defaultAction->setId(QStringLiteral("default"));
+
+    Q_EMIT defaultActionChanged();
+    if (d->id >= 0) {
+        d->updateTimer.start();
+    }
+
+    return d->defaultAction;
+}
+
+void KNotification::setDefaultActionQml(KNotificationAction *defaultAction)
 {
     if (defaultAction == d->defaultAction) {
         return;
@@ -228,37 +299,21 @@ void KNotification::setDefaultAction(const QString &defaultAction)
 
     d->needUpdate = true;
     d->defaultAction = defaultAction;
+    d->ownsActions = false;
+
+    d->defaultAction->setId(QStringLiteral("default"));
+
     Q_EMIT defaultActionChanged();
     if (d->id >= 0) {
         d->updateTimer.start();
     }
 }
 
-QString KNotification::defaultAction() const
+KNotificationAction *KNotification::defaultAction() const
 {
     return d->defaultAction;
 }
 
-KNotification::ContextList KNotification::contexts() const
-{
-    return d->contexts;
-}
-
-void KNotification::setContexts(const KNotification::ContextList &contexts)
-{
-    d->contexts = contexts;
-}
-
-void KNotification::addContext(const KNotification::Context &context)
-{
-    d->contexts << context;
-}
-
-void KNotification::addContext(const QString &context_key, const QString &context_value)
-{
-    d->contexts << qMakePair(context_key, context_value);
-}
-
 KNotification::NotificationFlags KNotification::flags() const
 {
     return d->flags;
@@ -321,30 +376,17 @@ void KNotification::setUrgency(Urgency urgency)
     }
 }
 
-void KNotification::activate(unsigned int action)
+void KNotification::activate(const QString &actionId)
 {
-    switch (action) {
-    case 0:
-#if KNOTIFICATIONS_BUILD_DEPRECATED_SINCE(5, 76)
-        Q_EMIT activated();
-#endif
-        Q_EMIT defaultActivated();
-        break;
-    case 1:
-        Q_EMIT action1Activated();
-        break;
-    case 2:
-        Q_EMIT action2Activated();
-        break;
-    case 3:
-        Q_EMIT action3Activated();
-        break;
+    if (d->defaultAction && actionId == QLatin1String("default")) {
+        Q_EMIT d->defaultAction->activated();
     }
 
-    // emitting activated() makes the Manager close all the active plugins
-    // which will deref() the KNotification object, which will result
-    // in closing the notification
-    Q_EMIT activated(action);
+    for (KNotificationAction *action : d->actions) {
+        if (action->id() == actionId) {
+            Q_EMIT action->activated();
+        }
+    }
 }
 
 void KNotification::close()
@@ -366,40 +408,6 @@ void KNotification::close()
     }
 }
 
-#if KNOTIFICATIONS_BUILD_DEPRECATED_SINCE(5, 67)
-void KNotification::raiseWidget()
-{
-    if (!d->widget) {
-        return;
-    }
-
-    d->Private::raiseWidget(d->widget);
-}
-#endif
-
-#if KNOTIFICATIONS_BUILD_DEPRECATED_SINCE(5, 67)
-void KNotification::Private::raiseWidget(QWidget *w)
-{
-    // TODO  this function is far from finished.
-    if (w->isTopLevel()) {
-        w->raise();
-#if HAVE_KWINDOWSYSTEM
-        if (!xdgActivationToken.isEmpty()) {
-            KWindowSystem::setCurrentXdgActivationToken(xdgActivationToken);
-        }
-        KWindowSystem::activateWindow(w->winId());
-#endif
-    } else {
-        QWidget *pw = w->parentWidget();
-        raiseWidget(pw);
-
-        if (QTabWidget *tab_widget = qobject_cast<QTabWidget *>(pw)) {
-            tab_widget->setCurrentIndex(tab_widget->indexOf(w));
-        }
-    }
-}
-#endif
-
 static QString defaultComponentName()
 {
 #if defined(Q_OS_ANDROID)
@@ -413,12 +421,10 @@ KNotification *KNotification::event(const QString &eventid,
                                     const QString &title,
                                     const QString &text,
                                     const QPixmap &pixmap,
-                                    QWidget *widget,
                                     const NotificationFlags &flags,
                                     const QString &componentName)
 {
     KNotification *notify = new KNotification(eventid, flags);
-    notify->setWidget(widget);
     notify->setTitle(title);
     notify->setText(text);
     notify->setPixmap(pixmap);
@@ -429,37 +435,30 @@ KNotification *KNotification::event(const QString &eventid,
     return notify;
 }
 
-KNotification *KNotification::event(const QString &eventid,
-                                    const QString &text,
-                                    const QPixmap &pixmap,
-                                    QWidget *widget,
-                                    const NotificationFlags &flags,
-                                    const QString &componentName)
-{
-    return event(eventid, QString(), text, pixmap, widget, flags, componentName);
-}
-
 KNotification *
-KNotification::event(StandardEvent eventid, const QString &title, const QString &text, const QPixmap &pixmap, QWidget *widget, const NotificationFlags &flags)
+KNotification::event(const QString &eventid, const QString &text, const QPixmap &pixmap, const NotificationFlags &flags, const QString &componentName)
 {
-    return event(standardEventToEventId(eventid), title, text, pixmap, widget, flags | DefaultEvent);
+    return event(eventid, QString(), text, pixmap, flags, componentName);
 }
 
-KNotification *KNotification::event(StandardEvent eventid, const QString &text, const QPixmap &pixmap, QWidget *widget, const NotificationFlags &flags)
+KNotification *KNotification::event(StandardEvent eventid, const QString &title, const QString &text, const QPixmap &pixmap, const NotificationFlags &flags)
 {
-    return event(eventid, QString(), text, pixmap, widget, flags);
+    return event(standardEventToEventId(eventid), title, text, pixmap, flags | DefaultEvent);
+}
+
+KNotification *KNotification::event(StandardEvent eventid, const QString &text, const QPixmap &pixmap, const NotificationFlags &flags)
+{
+    return event(eventid, QString(), text, pixmap, flags);
 }
 
 KNotification *KNotification::event(const QString &eventid,
                                     const QString &title,
                                     const QString &text,
                                     const QString &iconName,
-                                    QWidget *widget,
                                     const NotificationFlags &flags,
                                     const QString &componentName)
 {
     KNotification *notify = new KNotification(eventid, flags);
-    notify->setWidget(widget);
     notify->setTitle(title);
     notify->setText(text);
     notify->setIconName(iconName);
@@ -470,15 +469,14 @@ KNotification *KNotification::event(const QString &eventid,
     return notify;
 }
 
-KNotification *
-KNotification::event(StandardEvent eventid, const QString &title, const QString &text, const QString &iconName, QWidget *widget, const NotificationFlags &flags)
+KNotification *KNotification::event(StandardEvent eventid, const QString &title, const QString &text, const QString &iconName, const NotificationFlags &flags)
 {
-    return event(standardEventToEventId(eventid), title, text, iconName, widget, flags | DefaultEvent);
+    return event(standardEventToEventId(eventid), title, text, iconName, flags | DefaultEvent);
 }
 
-KNotification *KNotification::event(StandardEvent eventid, const QString &title, const QString &text, QWidget *widget, const NotificationFlags &flags)
+KNotification *KNotification::event(StandardEvent eventid, const QString &title, const QString &text, const NotificationFlags &flags)
 {
-    return event(standardEventToEventId(eventid), title, text, standardEventToIconName(eventid), widget, flags | DefaultEvent);
+    return event(standardEventToEventId(eventid), title, text, standardEventToIconName(eventid), flags | DefaultEvent);
 }
 
 void KNotification::ref()
@@ -495,9 +493,9 @@ void KNotification::deref()
     }
 }
 
-void KNotification::beep(const QString &reason, QWidget *widget)
+void KNotification::beep(const QString &reason)
 {
-    event(QStringLiteral("beep"), reason, QPixmap(), widget, CloseOnTimeout | DefaultEvent);
+    event(QStringLiteral("beep"), reason, QPixmap(), CloseOnTimeout | DefaultEvent);
 }
 
 void KNotification::sendEvent()
@@ -554,21 +552,6 @@ void KNotification::update()
     }
 }
 
-bool KNotification::eventFilter(QObject *watched, QEvent *event)
-{
-#ifdef QT_WIDGETS_LIB
-    if (watched == d->widget) {
-        if (event->type() == QEvent::WindowActivate) {
-            if (d->flags & CloseWhenWidgetActivated) {
-                QTimer::singleShot(500, this, &KNotification::close);
-            }
-        }
-    }
-#endif
-
-    return false;
-}
-
 QString KNotification::standardEventToEventId(KNotification::StandardEvent event)
 {
     QString eventId;
@@ -649,4 +632,27 @@ QString KNotification::xdgActivationToken() const
     return d->xdgActivationToken;
 }
 
+void KNotification::setWindow(QWindow *window)
+{
+    if (window == d->window) {
+        return;
+    }
+
+    disconnect(d->window, &QWindow::activeChanged, this, &KNotification::slotWindowActiveChanged);
+    d->window = window;
+    connect(d->window, &QWindow::activeChanged, this, &KNotification::slotWindowActiveChanged);
+}
+
+void KNotification::slotWindowActiveChanged()
+{
+    if (d->window->isActive() && (d->flags & CloseWhenWindowActivated)) {
+        close();
+    }
+}
+
+QWindow *KNotification::window() const
+{
+    return d->window;
+}
+
 #include "moc_knotification.cpp"
